
# This file was generated by 'versioneer.py' (0.14) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.52.0'
version_full = '18825058a6e32bb92f3941f3c8d14a818f19279f'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

